import 'dart:convert';
import 'package:dartz/dartz.dart';
import 'orm_utills.dart';
import 'orm_domain_types.dart';
import 'package:intl/intl.dart';
import 'package:data_db/data_db.dart';
import 'package:strings/strings.dart';
import 'orm_filter.dart';
import 'orm_filter_operator.dart';
import 'orm_ordenator.dart';

class Event {
  Function()? action;
  dynamic execute() => action!();
}

class NewRecordEvent {
  GenericDataFields Function()? action;
  GenericDataFields execute() => action!();
}

class Event_executeCommand {
  Future<Either<ErrorSqlResult, ExecuteSuccesSqlResult>> Function(
      String methodName, String sql)? action;
  Future<Either<ErrorSqlResult, ExecuteSuccesSqlResult>> execute(
          String methodName, String sql) =>
      action!(methodName, sql);
}

class Event_openQuery {
  Future<Either<ErrorSqlResult, SelectSuccesSqlResult>> Function(
      String methodName, String sql)? action;
  Future<Either<ErrorSqlResult, SelectSuccesSqlResult>> execute(
          String methodName, String sql) =>
      action!(methodName, sql);
}

class DatasetEvent {
  Function(Dataset dataset)? action;
  dynamic execute(Dataset dataset) => action!(dataset);
}

abstract class GenericField {
  GenericDataFields? _owner;
  final String _fieldName;
  final bool _notNullable;
  final bool _isPrimaryKey;
  late bool _hasDefault;
  late bool _isSerial;
  bool _assigned = false;
  bool _modefied = false;
  String? _internalProblem;
  // ignore: unused_field
  final Type _dbType = DynamicDataField;

  String get columnNameToSelect => _fieldName;
  bool get isRequired => _notNullable && !_hasDefault;
  bool get isProblem => isRequired && !_assigned;

  GenericField(
      this._owner, this._fieldName, this._notNullable, this._isPrimaryKey) {
    _hasDefault = false;
    _isSerial = false;
  }

  GenericField.clone(GenericField field)
      : _fieldName = field._fieldName,
        _notNullable = field._notNullable,
        _isPrimaryKey = field._isPrimaryKey,
        _hasDefault = field._hasDefault,
        _internalProblem = field._internalProblem;

  String _getSqlValue() => '';

  void _verify() {}
  void clear() {
    _assigned = false;
    _modefied = false;
  }

  void copy(GenericField field) {
    _internalProblem = field._internalProblem;
    _assigned = field._assigned;
    _modefied = field._modefied;
  }

  bool _trySetValue(dynamic value);
  bool _tryPutValue(void Function(dynamic value) fn);

  bool loadDataFromDataset(Dataset dataset) {
    return dataset
        .fieldByName(_fieldName)
        .fold(() => false, (a) => a.tryGetValue(_trySetValue));
  }

  bool setDateToDataset(Dataset dataset) {
    return dataset.fieldByName(_fieldName).fold(() => false, (a) {
      return _tryPutValue(a.setValue);
    });
  }

  String toJson() => '{$_fieldName: null}';

  Option<Map<String, Object>> toProto3();
}

abstract class IntegerField extends GenericField {
  int _value = 0;
  @override
  // ignore: unused_field
  final Type _dbType = IntegerDataField;

  IntegerField(GenericDataFields owner, String fieldName, bool notNullable,
      bool isPrimaryKey)
      : super(owner, fieldName, notNullable, isPrimaryKey);

  IntegerField.clone(IntegerField field) : super.clone(field) {
    setValue(field._value);
  }

  Option<int> getValue() => option(_assigned, _value);
  Option<int> get value => getValue();

  int setValue(int value) {
    _owner?._change();
    _value = value;
    _assigned = true;
    _modefied = true;
    return value;
  }

  void setNull() {
    _value = 0;
    _assigned = false;
    _modefied = true;
  }

  @override
  void copy(GenericField field) {
    super.copy(field);
    if (field is IntegerField) {
      _value = field._value;
    }
  }

  @override
  bool _trySetValue(dynamic value) {
    if (value is int) {
      return setValue(value) == _value;
    } else if (value == null) {
      setNull();
      return true;
    } else {
      return false;
    }
  }

  //set value(int v) => setValue(v);

  String toStringQuoted() => quotedStr(_value.toString());

  @override
  String _getSqlValue() => _value.toString();

  @override
  void clear() {
    super.clear();
    _value = 0;
  }

  @override
  bool _tryPutValue(void Function(dynamic value) fn) {
    return getValue().fold(() => false, (a) {
      fn(a);
      return true;
    });
  }

  @override
  String toJson() {
    return getValue().fold(() {
      return '${jsonEncode(_fieldName)}: ${jsonEncode(null)}';
    }, (a) {
      return '${jsonEncode(_fieldName)}: ${jsonEncode(a)}';
    });
  }

  @override
  Option<Map<String, Object>> toProto3() {
    return value.map((a) => {
          camelize(_fieldName, true): {'value': a}
        });
  }
}

abstract class IntegerField_WithDefault extends IntegerField {
  final int _defaultValue;

  int get defaultValue => _defaultValue;

  IntegerField_WithDefault(GenericDataFields owner, String fieldName,
      bool notNullable, bool isPrimaryKey, this._defaultValue)
      : super(owner, fieldName, notNullable, isPrimaryKey) {
    _hasDefault = true;
    _value = _defaultValue;
    _assigned = true;
  }

  IntegerField_WithDefault.clone(IntegerField_WithDefault field)
      : _defaultValue = field._defaultValue,
        super.clone(field) {
    setValue(field._value);
  }

  @override
  void clear() {
    super.clear();
    _value = _defaultValue;
  }
}

abstract class SerialField extends IntegerField {
  SerialField(GenericDataFields owner, String fieldName, bool notNullable,
      bool isPrimaryKey)
      : super(owner, fieldName, notNullable, isPrimaryKey) {
    _hasDefault = true;
    _isSerial = true;
  }

  SerialField.clone(SerialField field) : super.clone(field) {
    setValue(field._value);
  }
}

abstract class NumericField extends GenericField {
  @override
  // ignore: unused_field
  final Type _dbType = NumericDataField;
  double _value = 0;

  NumericField(GenericDataFields owner, String fieldName, bool notNullable,
      bool isPrimaryKey)
      : super(owner, fieldName, notNullable, isPrimaryKey);

  NumericField.clone(NumericField field) : super.clone(field) {
    setValue(field._value);
  }

  Option<double> getValue() => option(_assigned, _value);
  Option<double> get value => getValue();

  double setValue(double value) {
    _owner?._change();
    _value = value;
    _assigned = true;
    _modefied = true;
    return value;
  }

  @override
  void copy(GenericField field) {
    super.copy(field);
    if (field is NumericField) {
      _value = field._value;
    }
  }

  String toStringQuoted() => quotedStr(_value.toString());

  @override
  String _getSqlValue() => toStringQuoted();

  @override
  void clear() {
    super.clear();
    _value = _value;
  }

  @override
  bool _trySetValue(dynamic value) {
    if (value is double) {
      return setValue(value) == _value;
    } else if (value == null) {
      setNull();
      return true;
    } else {
      return false;
    }
  }

  @override
  bool _tryPutValue(void Function(dynamic value) fn) {
    return getValue().fold(() => false, (a) {
      fn(a);
      return true;
    });
  }

  void setNull() {
    _value = 0;
    _assigned = false;
    _modefied = true;
  }

  @override
  String toJson() {
    return getValue().fold(() {
      return '${jsonEncode(_fieldName)}: ${jsonEncode(null)}';
    }, (a) {
      return '${jsonEncode(_fieldName)}: ${jsonEncode(a)}';
    });
  }

  @override
  Option<Map<String, Object>> toProto3() {
    return value.map((a) => {
          camelize(_fieldName, true): {'value': a}
        });
  }
}

abstract class NumericField_WithDefault extends NumericField {
  final double _defaultValue;

  NumericField_WithDefault(GenericDataFields owner, String fieldName,
      bool notNullable, bool isPrimaryKey, this._defaultValue)
      : super(owner, fieldName, notNullable, isPrimaryKey) {
    _hasDefault = true;
    _value = _defaultValue;
    _assigned = true;
  }

  NumericField_WithDefault.clone(NumericField_WithDefault field)
      : _defaultValue = field._defaultValue,
        super.clone(field) {
    setValue(field._value);
  }

  @override
  void clear() {
    super.clear();
    _value = _defaultValue;
  }
}

abstract class StringField extends GenericField {
  @override
  // ignore: unused_field
  final Type _dbType = StringDataField;
  String _value = '';
  final int _maxLength;

  StringField(GenericDataFields owner, String fieldName, bool notNullable,
      bool isPrimaryKey, this._maxLength)
      : super(owner, fieldName, notNullable, isPrimaryKey);

  StringField.clone(StringField field)
      : _maxLength = field._maxLength,
        super.clone(field) {
    setValue(field._value);
  }

  Option<String> getValue() => option(
      (_assigned && (_maxLength == -1 || (_value.length <= _maxLength))),
      _value);
  Option<String> get value => getValue();

  String setValue(String value) {
    _owner?._change();
    _value = value;
    _assigned = true;
    _modefied = true;
    if ((_maxLength > 0) && (_value.length > _maxLength)) {
      _internalProblem =
          'O conteúdo campo ${quotedStr(_fieldName)} está maior do que o tamanho permitido($_maxLength).';
    }
    return value;
  }

  void setNull() {
    _value = '';
    _assigned = false;
    _modefied = true;
  }

  @override
  void copy(GenericField field) {
    super.copy(field);
    if (field is StringField) {
      _value = field._value;
    }
  }

  String toStringQuoted() => quotedStr(_value);

  @override
  String _getSqlValue() => _assigned ? toStringQuoted() : 'null';

  @override
  void clear() {
    super.clear();
    _value = _value;
  }

  @override
  bool _trySetValue(dynamic value) {
    if (value is String) {
      return setValue(value) == _value;
    } else if (value == null) {
      setNull();
      return true;
    } else {
      return false;
    }
  }

  @override
  bool _tryPutValue(void Function(dynamic value) fn) {
    return getValue().fold(() => false, (a) {
      fn(a);
      return true;
    });
  }

  @override
  String toJson() {
    return getValue().fold(() {
      return '${jsonEncode(_fieldName)}: ${jsonEncode(null)}';
    }, (a) {
      return '${jsonEncode(_fieldName)}: ${jsonEncode(a)}';
    });
  }

  @override
  Option<Map<String, Object>> toProto3() {
    return value.map((a) => {
          camelize(_fieldName, true): {'value': a}
        });
  }
}

abstract class StringField_WithDefault extends StringField {
  final String _defaultValue;

  StringField_WithDefault(GenericDataFields owner, String fieldName,
      bool notNullable, bool isPrimaryKey, int maxLength, this._defaultValue)
      : super(owner, fieldName, notNullable, isPrimaryKey, maxLength) {
    _hasDefault = true;
    _value = _defaultValue;
    _assigned = true;
  }

  StringField_WithDefault.clone(StringField_WithDefault field)
      : _defaultValue = field._defaultValue,
        super.clone(field) {
    setValue(field._value);
  }

  @override
  void clear() {
    super.clear();
    _value = _defaultValue;
  }
}

abstract class BooleanField extends GenericField {
  @override
  // ignore: unused_field
  final Type _dbType = BooleanDataField;
  bool _value = false;

  BooleanField(GenericDataFields owner, String fieldName, bool notNullable,
      bool isPrimaryKey)
      : super(owner, fieldName, notNullable, isPrimaryKey);

  BooleanField.clone(BooleanField field) : super.clone(field) {
    setValue(field._value);
  }

  Option<bool> getValue() => option(_assigned, _value);
  Option<bool> get value => getValue();

  bool setValue(bool value) {
    _owner?._change();
    _value = value;
    _assigned = true;
    _modefied = true;
    return value;
  }

  void setNull() {
    _value = false;
    _assigned = false;
    _modefied = true;
  }

  @override
  void copy(GenericField field) {
    super.copy(field);
    if (field is BooleanField) {
      _value = field._value;
    }
  }

  String toStringQuoted() => quotedStr(_value.toString());

  @override
  String _getSqlValue() => _value.toString();

  @override
  void clear() {
    super.clear();
    _value = _value;
  }

  @override
  bool _trySetValue(dynamic value) {
    if (value is bool) {
      return setValue(value) == _value;
    } else if (value == null) {
      setNull();
      return true;
    } else {
      return false;
    }
  }

  @override
  bool _tryPutValue(void Function(dynamic value) fn) {
    return getValue().fold(() => false, (a) {
      fn(a);
      return true;
    });
  }

  @override
  String toJson() {
    return getValue().fold(() {
      return '${jsonEncode(_fieldName)}: ${jsonEncode(null)}';
    }, (a) {
      return '${jsonEncode(_fieldName)}: ${jsonEncode(a)}';
    });
  }

  @override
  Option<Map<String, Object>> toProto3() {
    return value.map((a) => {
          camelize(_fieldName, true): {'value': a}
        });
  }
}

abstract class BooleanField_WithDefault extends BooleanField {
  final bool _defaultValue;

  BooleanField_WithDefault(GenericDataFields owner, String fieldName,
      bool notNullable, bool isPrimaryKey, this._defaultValue)
      : super(owner, fieldName, notNullable, isPrimaryKey) {
    _hasDefault = true;
    _value = _defaultValue;
    _assigned = true;
  }

  BooleanField_WithDefault.clone(BooleanField_WithDefault field)
      : _defaultValue = field._defaultValue,
        super.clone(field) {
    setValue(field._value);
  }

  @override
  void clear() {
    super.clear();
    _value = _defaultValue;
  }
}

abstract class DateField extends GenericField {
  @override
  // ignore: unused_field
  final Type _dbType = DateDataField;
  final String _dtFormat = 'yyyy-MM-dd';
  DateTime _value = DateTime(2020);

  DateField(GenericDataFields owner, String fieldName, bool notNullable,
      bool isPrimaryKey)
      : super(owner, fieldName, notNullable, isPrimaryKey);

  DateField.clone(DateField field) : super.clone(field) {
    setValue(field._value);
  }

  Option<DateTime> getValue() => option(_assigned, _value);
  Option<DateTime> get value => getValue();

  DateTime setValue(DateTime value) {
    _owner?._change();
    DateTime.utc(_value.year, _value.month, _value.day);
    _assigned = true;
    _modefied = true;
    return _value;
  }

  @override
  void copy(GenericField field) {
    super.copy(field);
    if (field is DateField) {
      _value = field._value;
    }
  }

  String toStringQuoted() => quotedStr(format(_dtFormat));

  String format(String format) {
    return DateFormat(format).format(_value);
  }

  @override
  String _getSqlValue() => toStringQuoted();

  @override
  void clear() {
    super.clear();
    _value = _value;
  }

  void setNull() {
    _value = DateTime(0);
    _assigned = false;
    _modefied = true;
  }

  @override
  bool _trySetValue(dynamic value) {
    if (value is DateTime) {
      return setValue(value) == _value;
    } else if (value == null) {
      setNull();
      return true;
    } else {
      return false;
    }
  }

  @override
  bool _tryPutValue(void Function(dynamic value) fn) {
    return getValue().fold(() => false, (a) {
      fn(a);
      return true;
    });
  }

  @override
  String toJson() {
    return getValue().fold(() {
      return '${jsonEncode(_fieldName)}: ${jsonEncode(null)}';
    }, (a) {
      return '${jsonEncode(_fieldName)}: ${jsonEncode(a.toString())}';
    });
  }

  @override
  Option<Map<String, Object>> toProto3() {
    return value.map((a) => {
          camelize(_fieldName, true): {'value': format(_dtFormat)}
        });
  }
}

abstract class DateField_WithDefault extends DateField {
  late final DateTime _defaultValue;

  DateField_WithDefault(GenericDataFields owner, String fieldName,
      bool notNullable, bool isPrimaryKey, DateTime defaultValue)
      : super(owner, fieldName, notNullable, isPrimaryKey) {
    _defaultValue =
        DateTime(defaultValue.year, defaultValue.month, defaultValue.day);
    _hasDefault = true;
    _value = _defaultValue;
    _assigned = true;
  }

  DateField_WithDefault.clone(DateField_WithDefault field)
      : _defaultValue = field._defaultValue,
        super.clone(field) {
    setValue(field._value);
  }
}

abstract class DateTimeField extends GenericField {
  @override
  // ignore: unused_field
  final Type _dbType = DateTimeDataField;
  DateTime _value = DateTime(2020);
  final String _dtFormat = 'yyyy-MM-dd HH:mm:ss';

  DateTimeField(GenericDataFields owner, String fieldName, bool notNullable,
      bool isPrimaryKey)
      : super(owner, fieldName, notNullable, isPrimaryKey);

  DateTimeField.clone(DateTimeField field) : super.clone(field) {
    setValue(field._value);
  }

  Option<DateTime> getValue() => option(_assigned, _value);
  Option<DateTime> get value => getValue();

  DateTime setValue(DateTime value) {
    _owner?._change();
    _value = value;
    _assigned = true;
    _modefied = true;
    return _value;
  }

  void setNull() {
    _value = DateTime(0);
    _assigned = false;
    _modefied = true;
  }

  @override
  void copy(GenericField field) {
    super.copy(field);
    if (field is DateTimeField) {
      _value = field._value;
    }
  }

  String toStringQuoted() => quotedStr(format(_dtFormat));

  String format(String format) {
    return DateFormat(format).format(_value);
  }

  @override
  String _getSqlValue() => toStringQuoted();

  @override
  bool _trySetValue(dynamic value) {
    if (value is DateTime) {
      return setValue(value) ==
          DateTime.utc(_value.year, _value.month, _value.day, _value.hour,
              _value.minute, _value.second);
    } else if (value == null) {
      setNull();
      return true;
    } else {
      return false;
    }
  }

  @override
  bool _tryPutValue(void Function(dynamic value) fn) {
    return getValue().fold(() => false, (a) {
      fn(a);
      return true;
    });
  }

  @override
  String toJson() {
    return getValue().fold(() {
      return '${jsonEncode(_fieldName)}: ${jsonEncode(null)}';
    }, (a) {
      return '${jsonEncode(_fieldName)}: ${jsonEncode(a.toString())}';
    });
  }

  @override
  Option<Map<String, Object>> toProto3() {
    return value.map((a) => {
          camelize(_fieldName, true): {'value': format(_dtFormat)}
        });
  }
}

abstract class DateTimeField_WithDefault extends DateTimeField {
  final DateTime _defaultValue;

  DateTimeField_WithDefault(GenericDataFields owner, String fieldName,
      bool notNullable, bool isPrimaryKey, this._defaultValue)
      : super(owner, fieldName, notNullable, isPrimaryKey) {
    _hasDefault = true;
    _value = _defaultValue;
    _assigned = true;
  }

  DateTimeField_WithDefault.clone(DateTimeField_WithDefault field)
      : _defaultValue = field._defaultValue,
        super.clone(field) {
    setValue(field._value);
  }
}

abstract class TimeField extends GenericField {
  @override
  // ignore: unused_field
  final Type _dbType = TimeDataField;
  DateTime _value = DateTime(2020);
  final String _dtFormat = 'HH:mm:ss';

  TimeField(GenericDataFields owner, String fieldName, bool notNullable,
      bool isPrimaryKey)
      : super(owner, fieldName, notNullable, isPrimaryKey);

  TimeField.clone(TimeField field) : super.clone(field) {
    setValue(field._value);
  }

  Option<DateTime> getValue() => option(_assigned, _value);
  Option<DateTime> get value => getValue();

  DateTime setValue(DateTime value) {
    _owner?._change();
    _value = DateTime.utc(
        0, 0, 0, value.hour, value.minute, value.second, value.millisecond);
    _assigned = true;
    _modefied = true;
    return _value;
  }

  void setNull() {
    _value = DateTime(0);
    _assigned = false;
    _modefied = true;
  }

  @override
  void copy(GenericField field) {
    super.copy(field);
    if (field is DateTimeField) {
      _value = field._value;
    }
  }

  String toStringQuoted() => quotedStr(format(_dtFormat));

  String format(String format) {
    return DateFormat(format).format(_value);
  }

  @override
  String _getSqlValue() => toStringQuoted();

  @override
  bool _trySetValue(dynamic value) {
    if (value is DateTime) {
      return setValue(value) == _value;
    } else if (value == null) {
      setNull();
      return true;
    } else {
      return false;
    }
  }

  @override
  bool _tryPutValue(void Function(dynamic value) fn) {
    return getValue().fold(() => false, (a) {
      fn(a);
      return true;
    });
  }

  @override
  String toJson() {
    return getValue().fold(() {
      return '${jsonEncode(_fieldName)}: ${jsonEncode(null)}';
    }, (a) {
      return '${jsonEncode(_fieldName)}: ${jsonEncode(a.toString())}';
    });
  }

  @override
  Option<Map<String, Object>> toProto3() {
    return value.map((a) => {
          camelize(_fieldName, true): {'value': format(_dtFormat)}
        });
  }
}

abstract class TimeField_WithDefault extends TimeField {
  final DateTime _defaultValue;

  TimeField_WithDefault(GenericDataFields owner, String fieldName,
      bool notNullable, bool isPrimaryKey, this._defaultValue)
      : super(owner, fieldName, notNullable, isPrimaryKey) {
    _hasDefault = true;
    _value = _defaultValue;
    _assigned = true;
  }

  TimeField_WithDefault.clone(TimeField_WithDefault field)
      : _defaultValue = field._defaultValue,
        super.clone(field) {
    setValue(field._value);
  }
}

enum DataFieldStatus { emptyMode, readyMode, insertMode, editMode }

GenericDataFields? getBackup(GenericDataFields origin) => origin._backup;
void setBackup(GenericDataFields origin, GenericDataFields backup) =>
    origin._backup = backup;

abstract class GenericDataFields {
  late GenericDataFields _backup;
  DataFieldStatus status = DataFieldStatus.emptyMode;
  late bool _assigned = false;
  late List<GenericField> fields;

  bool get isReady => status == DataFieldStatus.readyMode;
  bool get isNotready => !isReady;
  bool get assigned => _assigned;

  bool get inChange => ((status == DataFieldStatus.editMode) ||
      (status == DataFieldStatus.insertMode));
  bool get notInChange => !inChange;

  void _change() {
    if (notInChange) {
      if (_assigned) {
        try {
          backup();
          _backup._assigned = true;
          _backup.status = DataFieldStatus.readyMode;
          status = DataFieldStatus.editMode;
        } on Exception catch (e) {
          print(e.toString());
        }
      } else {
        status = DataFieldStatus.insertMode;
      }
    }
  }

  void backup() {}
  void restore() {}

  bool contains(List<GenericField> fields) {
    var x = 0;

    var pkfields = this.fields.where((element) => element._isPrimaryKey);

    pkfields.forEach((pkField) {
      var o = fields
          .indexWhere((element) => element._fieldName == pkField._fieldName);
      if (o >= 0) {
        var f = fields[o];
        if (f._getSqlValue() == pkField._getSqlValue()) {
          x++;
        }
      }
    });

    if (x == pkfields.length) {
      return true;
    } else {
      return false;
    }
  }

  void finalize() {
    _assigned = true;
    status = DataFieldStatus.readyMode;
  }
}

void setORMConnection(ORM orm, SqlConnection connection) =>
    orm._sqlConnection = connection;

abstract class ORM {
  late final SqlConnection _sqlConnection;
  final String _schemaName;
  final String _tableName;
  final List<GenericDataFields> _records = <GenericDataFields>[];
  int _rowIndex = 0;
  late final String _className =
      toString().replaceFirst('Instance of ', '').replaceAll("'", '');

  //events
  final NewRecordEvent _addRecord;

  int get rowIndex => _rowIndex;
  int get count => _records.where((element) => element.assigned).length;
  SqlConnection get sqlConnection => _sqlConnection;

  GenericDataFields get _currentRecord => _records[_rowIndex];

  String get _fullTableName => '$_schemaName.$_tableName';

  ORM(this._schemaName, this._tableName, this._addRecord);

  void _newREcord() {
    _records.add(_addRecord.execute());
  }

  void newRecord() {
    _newREcord();
    _rowIndex++;
  }

  String getFilter(List<OrmFilter> filters) {
    var result = '';
    filters.forEach((element) {
      result += element.getFilter();
    });
    if (result.substring(0, 5) == ' and ') {
      return result.replaceFirst(' and ', 'where ');
    } else if (result.substring(0, 4) == ' or ') {
      return result.replaceFirst(' or ', 'where ');
    } else {
      return 'where ' + result;
    }
  }

  String getSelectSql(OrmSelectableColumns columns, OrmSelectWhere where,
      GenericOrderBy orderBy,
      {int? limit, int? offset}) {
    var result =
        'select COUNT(*) OVER() qqqtttddd_rrreeeggg, ${columns.toString()} from $_fullTableName ${where.toString()}';
    var s = orderBy.toString();
    if (s != '') {
      result += ' $s';
    }

    if (limit != null) {
      result += ' limit $limit';
    }
    if (offset != null) {
      result += ' offset $offset';
    }

    return result;
  }

  Future<Either<ErrorSqlResult, ExecuteSuccesSqlResult>> insert() async {
    return (_prepareInsertSQL()).fold((l) => left(l..methodName = 'insert'),
        (r) async {
      try {
        return (await _sqlConnection.executeCommand('insert', r))
            .fold((l) => left(l), (r) {
          if (_currentRecord.assigned) {
            var x = _records.removeAt(_rowIndex);
            _records.add(x._backup);
            _records.add(x);
            _rowIndex++;
          }

          if (r.dataset != null) {
            _loadReturningValues(r.dataset as Dataset);
          }
          _currentRecord.finalize();

          return right(r);
        });
      } catch (e) {
        return left(ErrorSqlResult(_className, 'insert', '', [e.toString()]));
      }
    });
  }

  void _loadReturningValues(Dataset dataset) {
    _currentRecord.fields
        .where((element) => element._isSerial)
        .forEach((element) {
      dataset.tryGetValue(element._fieldName, (value) {
        return element._trySetValue(value);
      });
    });
  }

  Future<Either<ErrorSqlResult, ExecuteSuccesSqlResult>> insertFromProto3Json(
      Map<String, dynamic> proto3Json) async {
    clear();
    try {
      _readProto3Json(proto3Json);

      return insert();
    } on Exception catch (e) {
      return left(ErrorSqlResult('', '', '', [e.toString()]));
    }
  }

  void _readProto3Json(Map<String, dynamic> proto3Json) {
    proto3Json.forEach((key, value) {
      _currentRecord.fields
          .singleWhere((field) => camelize(field._fieldName, true) == key)
          ._trySetValue((value as Map<String, dynamic>)['value']);
    });
  }

  Future<Either<ErrorSqlResult, ExecuteSuccesSqlResult>> deleteFromProto3Json(
      Map<String, dynamic> proto3Json) async {
    clear();
    try {
      _readProto3Json(proto3Json);

      return delete();
    } on Exception catch (e) {
      return left(ErrorSqlResult('', '', '', [e.toString()]));
    }
  }

  Future<Either<ErrorSqlResult, ExecuteSuccesSqlResult>> updateFromProto3Json(
      Map<String, dynamic> proto3Json) async {
    clear();
    try {
      var pkfields = _currentRecord.fields
          .where((element) => element._isPrimaryKey)
          .toList();
      pkfields.forEach((pkfield) {
        if (proto3Json.containsKey(camelize(pkfield._fieldName, true))) {
          pkfield._trySetValue((proto3Json[camelize(pkfield._fieldName, true)]
              as Map<String, dynamic>)['value']);
        }
      });

      if (pkfields.length !=
          pkfields.where((element) => element._assigned).length) {
        return left(
            ErrorSqlResult('', '', '', ['All pkField must be assigned']));
      } else {
        return (await _materialize(pkfields)).fold((l) {
          return left(l);
        }, (r) {
          proto3Json.forEach((key, value) {
            _currentRecord.fields
                .singleWhere((field) => camelize(field._fieldName, true) == key)
                ._trySetValue((value as Map<String, dynamic>)['value']);
          });
          return update();
        });
      }
    } catch (e) {
      return left(ErrorSqlResult('', '', '', [e.toString()]));
    }
  }

  Future<Either<ErrorSqlResult, ExecuteSuccesSqlResult>> update() async {
    return (_prepareUpdateSQL()).fold((l) => left(l..methodName = 'update'),
        (sql) async {
      return (await _sqlConnection.executeCommand('update', sql))
          .fold((l) => left(l), (r) {
        if (r.rowsAffected == 1) {
          _currentRecord.finalize();
          return right(r);
        } else {
          return left(
              ErrorSqlResult(_className, 'update', sql, ['record not found']));
        }
      });
    });
  }

  Future<Either<ErrorSqlResult, ExecuteSuccesSqlResult>> delete() async {
    return (_prepareDeleteSQL()).fold((l) => left(l..methodName = 'delete'),
        (sql) async {
      return (await _sqlConnection.executeCommand('delete', sql))
          .fold((l) => left(l), (r) {
        if (r.rowsAffected == 1) {
          _records.removeAt(_rowIndex);
          if (_rowIndex == 0) {
            _newREcord();
          } else {
            _rowIndex--;
          }
          return right(r);
        } else {
          return left(
              ErrorSqlResult(_className, 'delete', sql, ['record not found']));
        }
      });
    });
  }

  Future<Either<ErrorSqlResult, ExecuteSuccesSqlResult>> _deleteRecord(
      List<GenericField> fields) async {
    return (_prepareDeleteREcordSQL(fields))
        .fold((l) => left(l..methodName = 'deleteRecord'), (sql) async {
      return (await _sqlConnection.executeCommand('deleteRecord', sql))
          .fold((l) => left(l), (r) {
        if (r.rowsAffected >= 1) {
          cancelAllChanges();
          var x = locateField(fields);
          if (x > -1) {
            _records.removeAt(x);
            if (_rowIndex > 0) {
              _rowIndex--;
            } else {
              _newREcord();
            }
          }
          return right(r);
        } else {
          return left(ErrorSqlResult(
              _className, 'deleteRecord', sql, ['record not found']));
        }
      });
    });
  }

  Future<Either<ErrorSqlResult, SelectSuccesSqlResult>> _materialize(
      List<GenericField> fields) async {
    if (locate(fields)) {
      return right(SelectSuccesSqlResult(_className, 'materialize', 1, []));
    } else {
      return (_prepareMaterializeSQL(fields)).fold((l) {
        return left(l..methodName = 'materialize');
      }, (sql) async {
        return (await _sqlConnection.openQuery('materialize', sql)).fold((l) {
          return left(l);
        }, (res) {
          if (res.dataset.count > 0) {
            if (count > 0) {
              _newREcord();
              _rowIndex++;
            }

            return loadFromDataset(res.dataset).fold((l) {
              return left(ErrorSqlResult(_className, 'materialize', '', [l]));
            }, (r) {
              return right(res);
            });
          } else {
            return left(ErrorSqlResult(
                _className, 'materialize', sql, ['record not found']));
          }
        });
      });
    }
  }

  Either<ErrorSqlResult, String> _prepareDeleteREcordSQL(
      List<GenericField> fields) {
    return getSQL(getDeleteSQL(fields, _fullTableName));
  }

  Either<ErrorSqlResult, String> _prepareMaterializeSQL(
      List<GenericField> fields) {
    return getSQL(getMaterializeSQL(fields, _fullTableName,
        columns: getColumnsToSelect(_currentRecord.fields)));
  }

  Either<ErrorSqlResult, String> _prepareInsertSQL() {
    return verifyAllFields(_currentRecord.fields).fold(
        (problemFields) => left(ErrorSqlResult(
            _className, '_prepareInsertSQL', null, problemFields)),
        (assignedFields) =>
            getSQL(getInsertSQL(_currentRecord.fields, _fullTableName)));
  }

  Either<ErrorSqlResult, String> _prepareUpdateSQL() {
    return verifyAllFields(_currentRecord.fields).fold(
        (problemFields) => left(ErrorSqlResult(
            _className, '_prepareUpdateSQL', null, problemFields)),
        (assignedFields) => getSQL(getUpdateSQL(
            getAssignedFields(_currentRecord.fields), _fullTableName)));
  }

  Either<ErrorSqlResult, String> _prepareDeleteSQL() {
    return verifyPKFields(_currentRecord.fields).fold(
        (problemFields) => left(ErrorSqlResult(
            _className, '_prepareDeleteSQL', null, problemFields)),
        (assignedFields) =>
            getSQL(getDeleteSQL(assignedFields, _fullTableName)));
  }

  Either<String, int> loadFromDataset(Dataset dataset) {
    if (dataset.isNotempty) {
      dataset.first();
      _currentRecord.fields.forEach((element) {
        element.loadDataFromDataset(dataset);
      });

      _currentRecord.finalize();

      return right(dataset.count);
    } else {
      return left('Dataset is empty.');
    }
  }

  void cancelChanges() {
    if ((_currentRecord.inChange) & (_currentRecord.assigned)) {
      _currentRecord.restore();
    }
  }

  void cancelAllChanges() {
    _records.forEach((element) {
      if ((element.inChange) & (element.assigned)) {
        element.restore();
      }
    });
  }

  void clear() {
    _records.clear();
    _newREcord();
    _rowIndex = 0;
  }

  bool locate(List<GenericField> fields) {
    var x = locateField(fields);
    if (x >= 0) {
      _rowIndex = x;
      return true;
    } else {
      return false;
    }
  }

  int locateField(List<GenericField> xfields) {
    var x = _records.indexWhere(
        (element) => element._assigned && element.contains(xfields));
    if (x >= 0) {
      if (_records[x]._assigned) {
        if (_records[x].inChange) {
          _records[x].restore();
        }

        return x;
      } else {
        return -1;
      }
    } else {
      return -1;
    }
  }

  String toJson() {
    var x = '';
    var r = '';
    cancelAllChanges();
    _records.where((element) => element.isReady).forEach((element) {
      r = '';
      element.fields.forEach((field) {
        r += ', ${field.toJson()}';
      });
      x += ', {${r.replaceFirst(', ', '')}}';
    });

    return '[${x.replaceFirst(', ', '')}]';
  }

  Map<String, Map<String, Object>> recordToProto3Json() {
    var data = <String, Object>{};

    _currentRecord.fields
        .where((element) => element._assigned)
        .forEach((field) {
      field.toProto3().map((a) => data.addAll(a));
    });
    return {'data': data};
  }
}

Either<String, List<GenericField>> _getPKFields(List<GenericField> fields) {
  var pk = fields.where((field) => field._isPrimaryKey);
  if (pk.isNotEmpty) {
    return right(pk.toList());
  } else {
    return left('No primary Key assigned.');
  }
}

Either<String, List<GenericField>> _getModefiedFields(
    List<GenericField> assignedFields) {
  var updatableFields =
      assignedFields.where((field) => !field._isPrimaryKey && field._modefied);
  if (updatableFields.isNotEmpty) {
    return right(updatableFields.toList());
  } else {
    return left('No modefied fields');
  }
}

String _getUpdateSyntaxFromFields(List<GenericField> fields) {
  var columns = '';

  fields.forEach((field) {
    columns += ', ${field._fieldName} = ${field._getSqlValue()}';
  });

  return columns.replaceFirst(', ', '');
}

String _getWhereSyntaxFromFields(List<GenericField> fields) {
  var primaryKey = '';

  fields.forEach((field) {
    primaryKey += 'and ${field._fieldName} = ${field._getSqlValue()}';
  });

  return primaryKey.replaceFirst('and ', '');
}

Either<String, String> getInsertSQL(
    List<GenericField> fields, String fullTableName) {
  var columns = '';
  var values = '';
  var returning = '';
  var sql = 'insert into $fullTableName';

  try {
    getAssignedFields(fields).forEach((field) {
      columns += ',${field._fieldName}';
      values += ',${field._getSqlValue()}';
    });

    fields.where((element) => element._isSerial).forEach((field) {
      returning += ',${field._fieldName}';
    });

    sql +=
        ' (${columns.replaceFirst(',', '')}) values(${values.replaceFirst(',', '')})';
    if (returning.isNotEmpty) {
      sql += ' returning ${returning.replaceFirst(',', '')}';
    }

    return right(sql);
  } on Exception catch (e) {
    return left('Unknown exception: $e');
  }
}

Either<List<String>, List<GenericField>> verifyFieldsInternal(
    List<GenericField> assignedFields) {
  assignedFields.forEach((field) {
    field._verify();
  });

  var problemFields = getFieldsWith_InternalProblems(assignedFields);

  if (problemFields.isNotEmpty) {
    return left(getMessageFromProblemFields(problemFields));
  } else {
    return right(assignedFields);
  }
}

Either<List<String>, List<GenericField>> verifyAllnotNullableFields(
    List<GenericField> fields) {
  var notNullables = getNotNullableFields(fields);

  if (notNullables.isNotEmpty) {
    var requireds = getNotAssignedFields(notNullables);
    if (requireds.isNotEmpty) {
      return left(getMessageForrequiredFields(requireds));
    } else {
      return right(getAssignedFields(fields));
    }
  } else {
    return right(fields);
  }
}

Either<List<String>, List<GenericField>> verifyAllFields(
    List<GenericField> fields) {
  return verifyAllnotNullableFields(fields)
      .fold((l) => left(l), (r) => verifyFieldsInternal(r));
}

Either<String, String> getDeleteSQL(
    List<GenericField> assignedFields, String fullTableName) {
  return right(
      'delete from $fullTableName where ${_getWhereSyntaxFromFields(assignedFields)}');
}

Either<String, String> getMaterializeSQL(
    List<GenericField> pkFields, String fullTableName,
    {String columns = '*'}) {
  return right(
      'select $columns from $fullTableName where ${_getWhereSyntaxFromFields(pkFields)}');
}

Either<List<String>, List<GenericField>> verifyPKFields(
    List<GenericField> fields) {
  return _getPKFields(fields)
      .fold((l) => left([l]), (r) => verifyAllPKFields(r));
}

Either<List<String>, List<GenericField>> verifyAllPKFields(
    List<GenericField> pkFields) {
  var requireds = getNotAssignedFields(pkFields);
  if (requireds.isNotEmpty) {
    return left(getMessageForrequiredFields(requireds));
  } else {
    return right(getAssignedFields(pkFields));
  }
}

Either<String, String> getUpdateSQL(
    List<GenericField> assignedFields, String tableName) {
  return _getPKFields(assignedFields).fold(
      (l) => left(l),
      (pk) => _getModefiedFields(assignedFields).fold((l) => left(l),
          (r) => right(_getMountedUpdateSql(pk, r, tableName))));
}

String _getMountedUpdateSql(
    List<GenericField> pk, List<GenericField> up, String tablename) {
  return 'update $tablename set ${_getUpdateSyntaxFromFields(up)}  where ${_getWhereSyntaxFromFields(pk)}';
}

String getColumnsToSelect(List<GenericField> fields) {
  var x = '';
  fields.forEach((field) {
    x += ', ${field.columnNameToSelect}';
  });

  return x.replaceFirst(', ', '');
}

List<GenericField> getNotNullableFields(List<GenericField> fields) {
  return fields.where((field) => field.isRequired).toList();
}

List<GenericField> getFieldsWith_InternalProblems(List<GenericField> fields) {
  return fields.where((field) => field._internalProblem != null).toList();
}

List<GenericField> getNotAssignedFields(List<GenericField> fields) {
  return fields.where((field) => !field._assigned).toList();
}

List<GenericField> getAssignedFields(List<GenericField> fields) {
  return fields.where((field) => (field._assigned || field._modefied)).toList();
}

List<String> getMessageForrequiredFields(List<GenericField> fields) {
  var x = [''];
  x.clear();

  fields.forEach((field) {
    x.add('O campo ${field._fieldName} é obrigatório');
  });

  return x;
}

List<String> getMessageFromProblemFields(List<GenericField> fields) {
  var x = [''];
  x.clear();

  fields.forEach((field) {
    x.add(field._internalProblem ?? '');
  });

  return x;
}

Either<ErrorSqlResult, String> getSQL(Either<String, String> sql) {
  return sql.fold(
      (l) => left(ErrorSqlResult('', 'getSQL', null, [l])), (r) => right(r));
}

Future<Either<ErrorSqlResult, ExecuteSuccesSqlResult>> getDeleteRecord(
    ORM orm, List<GenericField> fields) {
  return orm._deleteRecord(fields);
}

Future<Either<ErrorSqlResult, SelectSuccesSqlResult>> getMaterialize(
    ORM orm, List<GenericField> fields) {
  return orm._materialize(fields);
}

List<GenericDataFields> getRecords(ORM orm) => orm._records;

String getFieldName(GenericField field) => field._fieldName;

String getFilter(List<OrmFilter> filters,
    [List<GenericFilterOperator>? operators]) {
  var result = '';
  filters.forEach((element) {
    result += element.getFilter();
  });

  if (operators != null) {
    operators.forEach((element) {
      result += element.getSqlSyntax();
    });
  }
  if (result.isNotEmpty) {
    if (result.substring(0, 5) == ' and ') {
      return result.replaceFirst(' and ', 'where ');
    } else if (result.substring(0, 4) == ' or ') {
      return result.replaceFirst(' or ', 'where ');
    } else {
      return 'where ' + result;
    }
  } else {
    return '';
  }
}

abstract class OrmSelectableColumns {}

abstract class OrmSelectWhere {
  final _filters = <OrmFilter>[];
  final _operators = <GenericFilterOperator>[];

  List<OrmFilter> get filters => _filters;

  @override
  String toString() {
    return getFilter(filters, _operators);
  }

  Map<String, dynamic> toMap() {
    var operators = <Map<String, dynamic>>[];
    Map<String, dynamic> map;
    filters.where((element) => element.isNotEmpty).forEach((element) {
      map = element.toMap();
      if (map.containsKey('operators')) {
        (map['operators'] as List<Map<String, dynamic>>).forEach((element) {
          operators.add(element);
        });
      }
    });

    _operators.forEach((element) {
      operators.add(element.toMap());
    });
    return {'operators': operators};
  }

  void loadMap(Map<String, dynamic> map) {
    _operators.clear();
    if (map.containsKey('operators')) {
      var op = map['operators'];
      if (op is List<dynamic>) {
        op.forEach((element) {
          if (element is Map<String, dynamic>) {
            loadOperatorMap(_operators, element);
          }
        });
      }
    }
  }

  void loadFromJson(String json) {
    try {
      var map = jsonDecode(json);
      if (map is Map<String, dynamic>) {
        loadMap(map);
      }
    } catch (e) {
      //ignore
    }
  }

  String toJson() {
    return jsonEncode(toMap());
  }

  void clear() {
    _operators.clear();
    _filters.forEach((element) {
      element.clear();
    });
  }
}
