import 'package:static_postgres_orm/src/domain/orm_filter_operator.dart';

abstract class OrmFilter {
  final String _columnName;
  final _operators = <GenericFilterOperator>[];

  int get count => _operators.length;
  bool get isNotEmpty => count > 0;

  void clear() {
    _operators.clear();
  }

  String getFilter() {
    if (_operators.isNotEmpty) {
      var result = '';
      _operators.forEach((element) {
        result += element.getSqlSyntax();
      });
      if (result.isNotEmpty) {
        if (result.substring(1, 5) == ' and ') {
          return result.replaceFirst(' and ', '');
        } else {
          return result;
        }
      } else {
        return '';
      }
    } else {
      return '';
    }
  }

  Map<String, dynamic> toMap() {
    var result = <String, dynamic>{
      'name': runtimeType.toString(),
      'operators': <Map<String, dynamic>>[]
    };
    _operators.forEach((element) {
      (result['operators'] as List<dynamic>).add(element.toMap());
    });

    return result;
  }

  bool loadMap(Map<String, dynamic> map) {
    var result = false;
    if (map.containsKey('name')) {
      if (map['name'] == runtimeType.toString()) {
        if (map.containsKey('operators')) {
          var operators = map['operators'];

          if (operators is List<Map<String, dynamic>>) {
            operators.forEach((element) {
              loadOperatorMap(_operators, element);
            });

            result = true;
          }
        }
      }
    }

    return result;
  }

  OrmFilter(this._columnName);
}

abstract class StringFilter extends OrmFilter {
  void equalsTo(String value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(StringOperatorEqualsTo(_columnName, value, operation));
  }

  void notEqualsTo(String value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(StringOperatorNotEqualsTo(_columnName, value, operation));
  }

  void isIn(List<String> value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(StringOperatorIn(_columnName, value, operation));
  }

  void notIn(List<String> value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(StringOperatorNotIn(_columnName, value, operation));
  }

  void beginsWith(String value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(StringOperatorBeginsWith(_columnName, value, operation));
  }

  void notBeginsWith(String value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(StringOperatorNotBeginsWith(_columnName, value, operation));
  }

  void contains(String value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(StringOperatorContains(_columnName, value, operation));
  }

  void notContains(String value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(StringOperatorNotContains(_columnName, value, operation));
  }

  void endsWith(String value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(StringOperatorEndsWith(_columnName, value, operation));
  }

  void notEndsWith(String value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(StringOperatorNotEndsWith(_columnName, value, operation));
  }

  StringFilter(String columnName) : super(columnName);
}

abstract class IntegerFilter extends OrmFilter {
  void equalsTo(int value, [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(IntegerOperatorEqualsTo(_columnName, value, operation));
  }

  void notEqualsTo(int value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(IntegerOperatorNotEqualsTo(_columnName, value, operation));
  }

  void isIn(List<int> value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(IntegerOperatorIn(_columnName, value, operation));
  }

  void notIn(List<int> value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(IntegerOperatorNotIn(_columnName, value, operation));
  }

  void between(int initialValue, int finalValue,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(IntegerOperatorBetween(
        _columnName, initialValue, finalValue, operation));
  }

  void notBetween(int initialValue, int finalValue,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(IntegerOperatorNotBetween(
        _columnName, initialValue, finalValue, operation));
  }

  void moreThan(int value,
      [bool inclusive = false,
      LogicalOperator operation = LogicalOperator.and]) {
    _operators
        .add(IntegerOperatorMoreThan(_columnName, value, inclusive, operation));
  }

  void lessThan(int value,
      [bool inclusive = false,
      LogicalOperator operation = LogicalOperator.and]) {
    _operators
        .add(IntegerOperatorLessThan(_columnName, value, inclusive, operation));
  }

  IntegerFilter(String columnName) : super(columnName);
}

abstract class NumericFilter extends OrmFilter {
  void equalsTo(double value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(NumericOperatorEqualsTo(_columnName, value, operation));
  }

  void notEqualsTo(double value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(NumericOperatorNotEqualsTo(_columnName, value, operation));
  }

  void isIn(List<double> value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(NumericOperatorIn(_columnName, value, operation));
  }

  void notIn(List<double> value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(NumericOperatorNotIn(_columnName, value, operation));
  }

  void between(double initialValue, double finalValue,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(NumericOperatorBetween(
        _columnName, initialValue, finalValue, operation));
  }

  void notBetween(double initialValue, double finalValue,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(NumericOperatorNotBetween(
        _columnName, initialValue, finalValue, operation));
  }

  void moreThan(double value,
      [bool inclusive = false,
      LogicalOperator operation = LogicalOperator.and]) {
    _operators
        .add(NumericOperatorMoreThan(_columnName, value, inclusive, operation));
  }

  void lessThan(double value,
      [bool inclusive = false,
      LogicalOperator operation = LogicalOperator.and]) {
    _operators
        .add(NumericOperatorLessThan(_columnName, value, inclusive, operation));
  }

  NumericFilter(String columnName) : super(columnName);
}

abstract class DateFilter extends OrmFilter {
  void equalsTo(DateTime value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(DateOperatorEqualsTo(_columnName, value, operation));
  }

  void notEqualsTo(DateTime value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(DateOperatorNotEqualsTo(_columnName, value, operation));
  }

  void isIn(List<DateTime> value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(DateOperatorIn(_columnName, value, operation));
  }

  void notIn(List<DateTime> value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(DateOperatorNotIn(_columnName, value, operation));
  }

  void between(DateTime initialValue, DateTime finalValue,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(
        DateOperatorBetween(_columnName, initialValue, finalValue, operation));
  }

  void notBetween(DateTime initialValue, DateTime finalValue,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(DateOperatorNotBetween(
        _columnName, initialValue, finalValue, operation));
  }

  void moreThan(DateTime value,
      [bool inclusive = false,
      LogicalOperator operation = LogicalOperator.and]) {
    _operators
        .add(DateOperatorMoreThan(_columnName, value, inclusive, operation));
  }

  void lessThan(DateTime value,
      [bool inclusive = false,
      LogicalOperator operation = LogicalOperator.and]) {
    _operators
        .add(DateOperatorLessThan(_columnName, value, inclusive, operation));
  }

  DateFilter(String columnName) : super(columnName);
}

abstract class DateTimeFilter extends OrmFilter {
  void equalsTo(DateTime value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(DateTimeOperatorEqualsTo(_columnName, value, operation));
  }

  void notEqualsTo(DateTime value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(DateOperatorNotEqualsTo(_columnName, value, operation));
  }

  void isIn(List<DateTime> value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(DateOperatorIn(_columnName, value, operation));
  }

  void notIn(List<DateTime> value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(DateOperatorNotIn(_columnName, value, operation));
  }

  void between(DateTime initialValue, DateTime finalValue,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(
        DateOperatorBetween(_columnName, initialValue, finalValue, operation));
  }

  void notBetween(DateTime initialValue, DateTime finalValue,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(DateOperatorNotBetween(
        _columnName, initialValue, finalValue, operation));
  }

  void moreThan(DateTime value,
      [bool inclusive = false,
      LogicalOperator operation = LogicalOperator.and]) {
    _operators
        .add(DateOperatorMoreThan(_columnName, value, inclusive, operation));
  }

  void lessThan(DateTime value,
      [bool inclusive = false,
      LogicalOperator operation = LogicalOperator.and]) {
    _operators
        .add(DateOperatorLessThan(_columnName, value, inclusive, operation));
  }

  DateTimeFilter(String columnName) : super(columnName);
}

abstract class TimeFilter extends OrmFilter {
  void equalsTo(DateTime value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(TimeOperatorEqualsTo(_columnName, value, operation));
  }

  void notEqualsTo(DateTime value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(DateOperatorNotEqualsTo(_columnName, value, operation));
  }

  void isIn(List<DateTime> value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(DateOperatorIn(_columnName, value, operation));
  }

  void notIn(List<DateTime> value,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(DateOperatorNotIn(_columnName, value, operation));
  }

  void between(DateTime initialValue, DateTime finalValue,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(
        DateOperatorBetween(_columnName, initialValue, finalValue, operation));
  }

  void notBetween(DateTime initialValue, DateTime finalValue,
      [LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(DateOperatorNotBetween(
        _columnName, initialValue, finalValue, operation));
  }

  void moreThan(DateTime value,
      [bool inclusive = false,
      LogicalOperator operation = LogicalOperator.and]) {
    _operators
        .add(DateOperatorMoreThan(_columnName, value, inclusive, operation));
  }

  void lessThan(DateTime value,
      [bool inclusive = false,
      LogicalOperator operation = LogicalOperator.and]) {
    _operators
        .add(DateOperatorLessThan(_columnName, value, inclusive, operation));
  }

  TimeFilter(String columnName) : super(columnName);
}

abstract class BooleanFilter extends OrmFilter {
  void isTrue([LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(OperatorIsTrue(_columnName, operation));
  }

  void notIsTrue([LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(OperatorNotIsTrue(_columnName, operation));
  }

  void isFalse([LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(OperatorIsFalse(_columnName, operation));
  }

  void notIsFalse([LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(OperatorNotIsFalse(_columnName, operation));
  }

  BooleanFilter(String columnName) : super(columnName);
}

mixin NullableFilter on OrmFilter {
  void isNull([LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(IsNullOperator(_columnName, operation));
  }

  void isNotNull([LogicalOperator operation = LogicalOperator.and]) {
    _operators.add(IsNotNullOperator(_columnName, operation));
  }
}

void loadOperatorMap(
    List<GenericFilterOperator> operators, Map<String, dynamic> map) {
  if (map.containsKey('name')) {
    String name = map['name'];
    switch (name) {
      case 'StringOperatorEqualsTo':
        operators.add(StringOperatorEqualsTo.fromMap(map));
        break;
      case 'StringOperatorNotEqualsTo':
        operators.add(StringOperatorNotEqualsTo.fromMap(map));
        break;
      case 'StringOperatorIn':
        operators.add(StringOperatorIn.fromMap(map));
        break;
      case 'StringOperatorNotIn':
        operators.add(StringOperatorNotIn.fromMap(map));
        break;
      case 'StringOperatorBeginsWith':
        operators.add(StringOperatorBeginsWith.fromMap(map));
        break;
      case 'StringOperatorNotBeginsWith':
        operators.add(StringOperatorNotBeginsWith.fromMap(map));
        break;
      case 'StringOperatorContains':
        operators.add(StringOperatorContains.fromMap(map));
        break;
      case 'StringOperatorNotContains':
        operators.add(StringOperatorNotContains.fromMap(map));
        break;
      case 'StringOperatorEndsWith':
        operators.add(StringOperatorEndsWith.fromMap(map));
        break;
      case 'StringOperatorNotEndsWith':
        operators.add(StringOperatorNotEndsWith.fromMap(map));
        break;
      case 'IntegerOperatorEqualsTo':
        operators.add(IntegerOperatorEqualsTo.fromMap(map));
        break;
      case 'IntegerOperatorNotEqualsTo':
        operators.add(IntegerOperatorNotEqualsTo.fromMap(map));
        break;
      case 'IntegerOperatorIn':
        operators.add(IntegerOperatorIn.fromMap(map));
        break;
      case 'IntegerOperatorNotIn':
        operators.add(IntegerOperatorNotIn.fromMap(map));
        break;
      case 'IntegerOperatorBetween':
        operators.add(IntegerOperatorBetween.fromMap(map));
        break;
      case 'IntegerOperatorNotBetween':
        operators.add(IntegerOperatorNotBetween.fromMap(map));
        break;
      case 'IntegerOperatorMoreThan':
        operators.add(IntegerOperatorMoreThan.fromMap(map));
        break;
      case 'IntegerOperatorLessThan':
        operators.add(IntegerOperatorLessThan.fromMap(map));
        break;
      case 'NumericOperatorEqualsTo':
        operators.add(NumericOperatorEqualsTo.fromMap(map));
        break;
      case 'NumericOperatorNotEqualsTo':
        operators.add(NumericOperatorNotEqualsTo.fromMap(map));
        break;
      case 'NumericOperatorIn':
        operators.add(NumericOperatorIn.fromMap(map));
        break;
      case 'NumericOperatorNotIn':
        operators.add(NumericOperatorNotIn.fromMap(map));
        break;
      case 'NumericOperatorBetween':
        operators.add(NumericOperatorBetween.fromMap(map));
        break;
      case 'NumericOperatorNotBetween':
        operators.add(NumericOperatorNotBetween.fromMap(map));
        break;
      case 'NumericOperatorMoreThan':
        operators.add(NumericOperatorMoreThan.fromMap(map));
        break;
      case 'NumericOperatorLessThan':
        operators.add(NumericOperatorLessThan.fromMap(map));
        break;
      case 'OperatorIsTrue':
        operators.add(OperatorIsTrue.fromMap(map));
        break;
      case 'OperatorNotIsTrue':
        operators.add(OperatorNotIsTrue.fromMap(map));
        break;
      case 'OperatorIsFalse':
        operators.add(OperatorIsFalse.fromMap(map));
        break;
      case 'OperatorNotIsFalse':
        operators.add(OperatorNotIsFalse.fromMap(map));
        break;
      case 'DateOperatorEqualsTo':
        operators.add(DateOperatorEqualsTo.fromMap(map));
        break;
      case 'DateOperatorNotEqualsTo':
        operators.add(DateOperatorNotEqualsTo.fromMap(map));
        break;
      case 'DateOperatorIn':
        operators.add(DateOperatorIn.fromMap(map));
        break;
      case 'DateOperatorNotIn':
        operators.add(DateOperatorNotIn.fromMap(map));
        break;
      case 'DateOperatorBetween':
        operators.add(DateOperatorBetween.fromMap(map));
        break;
      case 'DateOperatorNotBetween':
        operators.add(DateOperatorNotBetween.fromMap(map));
        break;
      case 'DateOperatorMoreThan':
        operators.add(DateOperatorMoreThan.fromMap(map));
        break;
      case 'DateOperatorLessThan':
        operators.add(DateOperatorLessThan.fromMap(map));
        break;
      case 'DateTimeOperatorEqualsTo':
        operators.add(DateTimeOperatorEqualsTo.fromMap(map));
        break;
      case 'DateTimeOperatorNotEqualsTo':
        operators.add(DateTimeOperatorNotEqualsTo.fromMap(map));
        break;
      case 'DateTimeOperatorIn':
        operators.add(DateTimeOperatorIn.fromMap(map));
        break;
      case 'DateTimeOperatorNotIn':
        operators.add(DateTimeOperatorNotIn.fromMap(map));
        break;
      case 'DateTimeOperatorBetween':
        operators.add(DateTimeOperatorBetween.fromMap(map));
        break;
      case 'DateTimeOperatorNotBetween':
        operators.add(DateTimeOperatorNotBetween.fromMap(map));
        break;
      case 'DateTimeOperatorMoreThan':
        operators.add(DateTimeOperatorMoreThan.fromMap(map));
        break;
      case 'DateTimeOperatorLessThan':
        operators.add(DateTimeOperatorLessThan.fromMap(map));
        break;
      case 'TimeOperatorEqualsTo':
        operators.add(TimeOperatorEqualsTo.fromMap(map));
        break;
      case 'TimeOperatorNotEqualsTo':
        operators.add(TimeOperatorNotEqualsTo.fromMap(map));
        break;
      case 'TimeOperatorIn':
        operators.add(TimeOperatorIn.fromMap(map));
        break;
      case 'TimeOperatorNotIn':
        operators.add(TimeOperatorNotIn.fromMap(map));
        break;
      case 'TimeOperatorBetween':
        operators.add(TimeOperatorBetween.fromMap(map));
        break;
      case 'TimeOperatorNotBetween':
        operators.add(TimeOperatorNotBetween.fromMap(map));
        break;
      case 'TimeOperatorMoreThan':
        operators.add(TimeOperatorMoreThan.fromMap(map));
        break;
      case 'TimeOperatorLessThan':
        operators.add(TimeOperatorLessThan.fromMap(map));
        break;
      case 'IsNullOperator':
        operators.add(IsNullOperator.fromMap(map));
        break;
      case 'IsNotNullOperator':
        operators.add(IsNotNullOperator.fromMap(map));
        break;
    }
  }
}
