import 'package:dartz/dartz.dart';
//import 'package:static_postgres_orm/src/dependencies/postgresql/lib/postgres.dart';
import 'package:data_db/data_db.dart';
//import 'package:static_postgres_orm/src/dependencies/postgresql/lib/src/query.dart';
import '../domain/orm_data_types.dart';
import '../domain/orm_domain_types.dart';
import 'package:postgres/postgres.dart';

ExecuteSuccesSqlResult pgQueryToExecuteSuccesSqlResult(String className,
    String methodName, String sqlCommand, PostgreSQLResult pgQuery) {
  return ExecuteSuccesSqlResult(className, methodName, pgQuery.affectedRowCount)
    ..dataset = pgSqlToDataset(pgQuery);
}

SelectSuccesSqlResult pgQueryToSelectSuccesSqlResult(String className,
    String methodName, String sqlCommand, PostgreSQLResult pgQuery) {
  return SelectSuccesSqlResult(
      className, methodName, pgQuery.affectedRowCount, pgQuery)
    ..dataset = pgSqlToDataset(pgQuery);
}

ErrorSqlResult pgExceptionToErrorSqlResult(String className, String methodName,
    String sqlCommand, PostgreSQLException e) {
  return ErrorSqlResult(className, methodName, sqlCommand, [e.message ?? ''],
      errorCode: e.code);
}

class Postgres_SqlConnection extends SqlConnection {
  PostgreSQLExecutionContext? _connection;
  late String className;
  String host = '';
  int port = 5432;
  String databaseName = 'postgres';
  String username = 'postgres';
  String password = '';
  int timeoutInSeconds = 30;
  int queryTimeoutInSeconds = 30;
  String timeZone = 'UTC';
  bool useSSL = false;
  bool isUnixSocket = false;

  Postgres_SqlConnection(
    this.host,
    this.port,
    this.databaseName,
    this.username,
    this.password, {
    timeoutInSeconds,
    queryTimeoutInSeconds,
    timeZone,
    useSSL,
    isUnixSocket,
  }) {
    if (useSSL != null) this.useSSL = useSSL;
  }

  Postgres_SqlConnection.fromConnection(PostgreSQLExecutionContext con) {
    _connection = con;
  }

  Postgres_SqlConnection.loadConnection(PostgreSQLConnection con) {
    host = con.host;
    port = con.port;
    databaseName = con.databaseName;
    if (con.username != null) {
      username = con.username as String;
    }

    if (con.password != null) {
      password = con.password as String;
    }

    timeoutInSeconds = con.timeoutInSeconds;
    queryTimeoutInSeconds = con.queryTimeoutInSeconds;
    timeZone = con.timeZone;
    useSSL = con.useSSL;
    isUnixSocket = con.isUnixSocket;
  }

  PostgreSQLExecutionContext newConnection() {
    return _connection ??
        PostgreSQLConnection(host, port, databaseName,
            username: username,
            password: password,
            timeoutInSeconds: timeoutInSeconds,
            queryTimeoutInSeconds: queryTimeoutInSeconds,
            timeZone: timeZone,
            useSSL: useSSL,
            isUnixSocket: isUnixSocket);
  }

  @override
  Future<Either<ErrorSqlResult, ExecuteSuccesSqlResult>> executeCommand(
      String methodName, String sql) async {
    var _dbConnection = newConnection();
    try {
      if ((_dbConnection is PostgreSQLConnection) && (_dbConnection.isClosed)) {
        await _dbConnection.open();
      }

      return right(pgQueryToExecuteSuccesSqlResult(
          className, methodName, sql, (await _dbConnection.query(sql))));
    } on PostgreSQLException catch (e) {
      try {
        _dbConnection.cancelTransaction(reason: e.toString());
      } on Exception {
        // ignore
      }

      return left(pgExceptionToErrorSqlResult(className, methodName, sql, e));
    } catch (e) {
      try {
        _dbConnection.cancelTransaction(reason: e.toString());
      } on Exception {
        // ignore
      }
      return left(ErrorSqlResult(className, methodName, sql, [e.toString()]));
    } finally {
      if ((_dbConnection is PostgreSQLConnection) &&
          (!_dbConnection.isClosed)) {
        await _dbConnection.close();
      }
    }
  }

  @override
  Future<Either<ErrorSqlResult, SelectSuccesSqlResult>> openQuery(
      String methodName, String sql) async {
    var _dbConnection = newConnection();
    try {
      if ((_dbConnection is PostgreSQLConnection) && (_dbConnection.isClosed)) {
        await _dbConnection.open();
      }
      return right(pgQueryToSelectSuccesSqlResult(
          className, methodName, sql, (await _dbConnection.query(sql))));
    } on PostgreSQLException catch (e) {
      try {
        _dbConnection.cancelTransaction(reason: e.toString());
      } on Exception {
        // ignore
      }
      return left(pgExceptionToErrorSqlResult(className, methodName, sql, e));
    } catch (e) {
      try {
        _dbConnection.cancelTransaction(reason: e.toString());
      } on Exception {
        // ignore
      }
      return left(ErrorSqlResult(className, methodName, sql, [e.toString()]));
    } finally {
      if ((_dbConnection is PostgreSQLConnection) &&
          (!_dbConnection.isClosed)) {
        await _dbConnection.close();
      }
    }
  }
}

abstract class PostgressORM extends ORM {
  late final String _className =
      toString().replaceFirst('Instance of ', '').replaceAll("'", '');
  late final Postgres_SqlConnection _postgresConnection;

  PostgressORM(
      this._postgresConnection, schemaName, tableName, NewRecordEvent addRecord)
      : super(schemaName, tableName, addRecord) {
    setORMConnection(this, _postgresConnection);
    _postgresConnection.className = _className;
  }

  PostgressORM.fromConnection(PostgreSQLExecutionContext _pgConnection,
      schemaName, tableName, NewRecordEvent addRecord)
      : super(schemaName, tableName, addRecord) {
    _postgresConnection = Postgres_SqlConnection.fromConnection(_pgConnection);
    setORMConnection(this, _postgresConnection);
    _postgresConnection.className = _className;
  }
}

DataField getFieldClass(Dataset ds, ColumnDescription field) {
  if (field.typeId == 16) {
    return BooleanDataField();
  } else if ((field.typeId == 16) ||
      (field.typeId == 19) ||
      (field.typeId == 1043) ||
      (field.typeId == 25)) {
    return StringDataField();
  } else if (field.typeId == 16) {
    return StringDataField();
  } else if ((field.typeId == 20) ||
      (field.typeId == 21) ||
      (field.typeId == 23)) {
    return IntegerDataField();
  } else if ((field.typeId == 700) ||
      (field.typeId == 701) ||
      (field.typeId == 790) ||
      (field.typeId == 1700)) {
    return NumericDataField();
  } else if (field.typeId == 1082) {
    return DateDataField();
  } else if ((field.typeId == 1083) || (field.typeId == 1266)) {
    return TimeDataField();
  } else if ((field.typeId == 1114) || (field.typeId == 1184)) {
    return DateTimeDataField();
  } else {
    return DynamicDataField();
  }
}

Type getFieldType(ColumnDescription field) {
  if (field.typeId == 16) {
    return BooleanDataField;
  } else if ((field.typeId == 16) ||
      (field.typeId == 19) ||
      (field.typeId == 1043) ||
      (field.typeId == 25)) {
    return StringDataField;
  } else if (field.typeId == 16) {
    return StringDataField;
  } else if ((field.typeId == 20) ||
      (field.typeId == 21) ||
      (field.typeId == 23)) {
    return IntegerDataField;
  } else if ((field.typeId == 700) ||
      (field.typeId == 701) ||
      (field.typeId == 790) ||
      (field.typeId == 1700)) {
    return NumericDataField;
  } else if (field.typeId == 1082) {
    return DateDataField;
  } else if ((field.typeId == 1083) || (field.typeId == 1266)) {
    return TimeDataField;
  } else if ((field.typeId == 1114) || (field.typeId == 1184)) {
    return DateTimeDataField;
  } else {
    return DynamicDataField;
  }
}

Dataset pgSqlToDataset(PostgreSQLResult pg) {
  var ds = Dataset();
  var fieldDescription = pg.columnDescriptions;
  fieldDescription.forEach((element) {
    ds.structure.addAll({element.columnName: getFieldType(element)});
  });

  var i = -1;

  for (final row in pg) {
    ds.newRecord();
    i = -1;
    fieldDescription.forEach((element) {
      i++;
      ds.trySetValue(element.columnName, row[i]);
    });
  }
  ds.first();
  return ds;
}
